from .lazyClassifier import LazyClassifier
from .lazyRegressor import LazyRegressor
from .lazyMTS import LazyMTS
from .lazydeepClassifier import LazyDeepClassifier
from .lazydeepRegressor import LazyDeepRegressor

__all__ = ["LazyClassifier", "LazyRegressor", "LazyMTS", 
"LazyDeepClassifier", "LazyDeepRegressor"]